@@ -22,57 +22,20 @@ use std::io::SeekFrom;
 use std::path::PathBuf;
 use std::sync::Arc;
 
-use futures::AsyncWriteExt;
 use log::debug;
-use serde::Deserialize;
-use serde::Serialize;
 use uuid::Uuid;
 
+use super::delete::HdfsDeleter;
 use super::lister::HdfsLister;
+use super::reader::HdfsReader;
 use super::writer::HdfsWriter;
 use crate::raw::*;
-use crate::services::hdfs::reader::HdfsReader;
+use crate::services::HdfsConfig;
 use crate::*;
 
-/// [Hadoop Distributed File System (HDFS™)](https://hadoop.apache.org/) support.
-///
-/// Config for Hdfs services support.
-#[derive(Default, Serialize, Deserialize, Clone, PartialEq, Eq)]
-#[serde(default)]
-#[non_exhaustive]
-pub struct HdfsConfig {
-    /// work dir of this backend
-    pub root: Option<String>,
-    /// name node of this backend
-    pub name_node: Option<String>,
-    /// kerberos_ticket_cache_path of this backend
-    pub kerberos_ticket_cache_path: Option<String>,
-    /// user of this backend
-    pub user: Option<String>,
-    /// enable the append capacity
-    pub enable_append: bool,
-    /// atomic_write_dir of this backend
-    pub atomic_write_dir: Option<String>,
-}
-
-impl Debug for HdfsConfig {
-    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
-        f.debug_struct("HdfsConfig")
-            .field("root", &self.root)
-            .field("name_node", &self.name_node)
-            .field(
-                "kerberos_ticket_cache_path",
-                &self.kerberos_ticket_cache_path,
-            )
-            .field("user", &self.user)
-            .field("enable_append", &self.enable_append)
-            .field("atomic_write_dir", &self.atomic_write_dir)
-            .finish_non_exhaustive()
-    }
-}
-
 impl Configurator for HdfsConfig {
-    fn into_builder(self) -> impl Builder {
+    type Builder = HdfsBuilder;
+    fn into_builder(self) -> Self::Builder {
         HdfsBuilder { config: self }
     }
 }
@@ -230,9 +193,9 @@ fn tmp_file_of(path: &str) -> String {
 /// Backend for hdfs services.
 #[derive(Debug, Clone)]
 pub struct HdfsBackend {
-    root: String,
+    pub root: String,
     atomic_write_dir: Option<String>,
-    client: Arc<hdrs::Client>,
+    pub client: Arc<hdrs::Client>,
     enable_append: bool,
 }
 
@@ -244,9 +207,11 @@ impl Access for HdfsBackend {
     type Reader = HdfsReader<hdrs::AsyncFile>;
     type Writer = HdfsWriter<hdrs::AsyncFile>;
     type Lister = Option<HdfsLister>;
+    type Deleter = oio::OneShotDeleter<HdfsDeleter>;
     type BlockingReader = HdfsReader<hdrs::File>;
     type BlockingWriter = HdfsWriter<hdrs::File>;
     type BlockingLister = Option<HdfsLister>;
+    type BlockingDeleter = oio::OneShotDeleter<HdfsDeleter>;
 
     fn info(&self) -> Arc<AccessorInfo> {
         let mut am = AccessorInfo::default();
@@ -254,6 +219,8 @@ impl Access for HdfsBackend {
             .set_root(&self.root)
             .set_native_capability(Capability {
                 stat: true,
+                stat_has_content_length: true,
+                stat_has_last_modified: true,
 
                 read: true,
 
@@ -264,10 +231,14 @@ impl Access for HdfsBackend {
                 delete: true,
 
                 list: true,
+                list_has_content_length: true,
+                list_has_last_modified: true,
 
                 rename: true,
                 blocking: true,
 
+                shared: true,
+
                 ..Default::default()
             });
 
@@ -327,45 +298,35 @@ impl Access for HdfsBackend {
     }
 
     async fn write(&self, path: &str, op: OpWrite) -> Result<(RpWrite, Self::Writer)> {
-        let (target_path, tmp_path) = if let Some(atomic_write_dir) = &self.atomic_write_dir {
-            let target_path = build_rooted_abs_path(&self.root, path);
-            let tmp_path = build_rooted_abs_path(atomic_write_dir, &tmp_file_of(path));
-
-            // If the target file exists, we should append to the end of it directly.
-            if op.append() && self.client.metadata(&target_path).is_ok() {
-                (target_path, None)
-            } else {
-                (target_path, Some(tmp_path))
+        let target_path = build_rooted_abs_path(&self.root, path);
+        let target_exists = match self.client.metadata(&target_path) {
+            Ok(_) => true,
+            Err(err) => {
+                if err.kind() != io::ErrorKind::NotFound {
+                    return Err(new_std_io_error(err));
+                }
+                false
             }
-        } else {
-            let p = build_rooted_abs_path(&self.root, path);
-            (p, None)
         };
 
-        if let Err(err) = self.client.metadata(&target_path) {
-            // Early return if other error happened.
-            if err.kind() != io::ErrorKind::NotFound {
-                return Err(new_std_io_error(err));
+        let should_append = op.append() && target_exists;
+        let tmp_path = self.atomic_write_dir.as_ref().and_then(|atomic_write_dir| {
+            // If the target file exists, we should append to the end of it directly.
+            if should_append {
+                None
+            } else {
+                Some(build_rooted_abs_path(atomic_write_dir, &tmp_file_of(path)))
             }
+        });
 
+        if !target_exists {
             let parent = get_parent(&target_path);
-
             self.client.create_dir(parent).map_err(new_std_io_error)?;
-
-            let mut f = self
-                .client
-                .open_file()
-                .create(true)
-                .write(true)
-                .async_open(&target_path)
-                .await
-                .map_err(new_std_io_error)?;
-            f.close().await.map_err(new_std_io_error)?;
         }
 
         let mut open_options = self.client.open_file();
         open_options.create(true);
-        if op.append() {
+        if should_append {
             open_options.append(true);
         } else {
             open_options.write(true);
@@ -378,35 +339,21 @@ impl Access for HdfsBackend {
 
         Ok((
             RpWrite::new(),
-            HdfsWriter::new(target_path, tmp_path, f, Arc::clone(&self.client)),
+            HdfsWriter::new(
+                target_path,
+                tmp_path,
+                f,
+                Arc::clone(&self.client),
+                target_exists,
+            ),
         ))
     }
 
-    async fn delete(&self, path: &str, _: OpDelete) -> Result<RpDelete> {
-        let p = build_rooted_abs_path(&self.root, path);
-
-        let meta = self.client.metadata(&p);
-
-        if let Err(err) = meta {
-            return if err.kind() == io::ErrorKind::NotFound {
-                Ok(RpDelete::default())
-            } else {
-                Err(new_std_io_error(err))
-            };
-        }
-
-        // Safety: Err branch has been checked, it's OK to unwrap.
-        let meta = meta.ok().unwrap();
-
-        let result = if meta.is_dir() {
-            self.client.remove_dir(&p)
-        } else {
-            self.client.remove_file(&p)
-        };
-
-        result.map_err(new_std_io_error)?;
-
-        Ok(RpDelete::default())
+    async fn delete(&self) -> Result<(RpDelete, Self::Deleter)> {
+        Ok((
+            RpDelete::default(),
+            oio::OneShotDeleter::new(HdfsDeleter::new(Arc::new(self.clone()))),
+        ))
     }
 
     async fn list(&self, path: &str, _: OpList) -> Result<(RpList, Self::Lister)> {
@@ -423,7 +370,7 @@ impl Access for HdfsBackend {
             }
         };
 
-        let rd = HdfsLister::new(&self.root, f);
+        let rd = HdfsLister::new(&self.root, f, path);
 
         Ok((RpList::default(), Some(rd)))
     }
@@ -526,43 +473,35 @@ impl Access for HdfsBackend {
     }
 
     fn blocking_write(&self, path: &str, op: OpWrite) -> Result<(RpWrite, Self::BlockingWriter)> {
-        let (target_path, tmp_path) = if let Some(atomic_write_dir) = &self.atomic_write_dir {
-            let target_path = build_rooted_abs_path(&self.root, path);
-            let tmp_path = build_rooted_abs_path(atomic_write_dir, &tmp_file_of(path));
-
-            // If the target file exists, we should append to the end of it directly.
-            if op.append() && self.client.metadata(&target_path).is_ok() {
-                (target_path, None)
-            } else {
-                (target_path, Some(tmp_path))
+        let target_path = build_rooted_abs_path(&self.root, path);
+        let target_exists = match self.client.metadata(&target_path) {
+            Ok(_) => true,
+            Err(err) => {
+                if err.kind() != io::ErrorKind::NotFound {
+                    return Err(new_std_io_error(err));
+                }
+                false
             }
-        } else {
-            let p = build_rooted_abs_path(&self.root, path);
-
-            (p, None)
         };
 
-        if let Err(err) = self.client.metadata(&target_path) {
-            // Early return if other error happened.
-            if err.kind() != io::ErrorKind::NotFound {
-                return Err(new_std_io_error(err));
+        let should_append = op.append() && target_exists;
+        let tmp_path = self.atomic_write_dir.as_ref().and_then(|atomic_write_dir| {
+            // If the target file exists, we should append to the end of it directly.
+            if should_append {
+                None
+            } else {
+                Some(build_rooted_abs_path(atomic_write_dir, &tmp_file_of(path)))
             }
+        });
 
+        if !target_exists {
             let parent = get_parent(&target_path);
-
             self.client.create_dir(parent).map_err(new_std_io_error)?;
-
-            self.client
-                .open_file()
-                .create(true)
-                .write(true)
-                .open(&target_path)
-                .map_err(new_std_io_error)?;
         }
 
         let mut open_options = self.client.open_file();
         open_options.create(true);
-        if op.append() {
+        if should_append {
             open_options.append(true);
         } else {
             open_options.write(true);
@@ -574,35 +513,21 @@ impl Access for HdfsBackend {
 
         Ok((
             RpWrite::new(),
-            HdfsWriter::new(target_path, tmp_path, f, Arc::clone(&self.client)),
+            HdfsWriter::new(
+                target_path,
+                tmp_path,
+                f,
+                Arc::clone(&self.client),
+                target_exists,
+            ),
         ))
     }
 
-    fn blocking_delete(&self, path: &str, _: OpDelete) -> Result<RpDelete> {
-        let p = build_rooted_abs_path(&self.root, path);
-
-        let meta = self.client.metadata(&p);
-
-        if let Err(err) = meta {
-            return if err.kind() == io::ErrorKind::NotFound {
-                Ok(RpDelete::default())
-            } else {
-                Err(new_std_io_error(err))
-            };
-        }
-
-        // Safety: Err branch has been checked, it's OK to unwrap.
-        let meta = meta.ok().unwrap();
-
-        let result = if meta.is_dir() {
-            self.client.remove_dir(&p)
-        } else {
-            self.client.remove_file(&p)
-        };
-
-        result.map_err(new_std_io_error)?;
-
-        Ok(RpDelete::default())
+    fn blocking_delete(&self) -> Result<(RpDelete, Self::BlockingDeleter)> {
+        Ok((
+            RpDelete::default(),
+            oio::OneShotDeleter::new(HdfsDeleter::new(Arc::new(self.clone()))),
+        ))
     }
 
     fn blocking_list(&self, path: &str, _: OpList) -> Result<(RpList, Self::BlockingLister)> {
@@ -619,7 +544,7 @@ impl Access for HdfsBackend {
             }
         };
 
-        let rd = HdfsLister::new(&self.root, f);
+        let rd = HdfsLister::new(&self.root, f, path);
 
         Ok((RpList::default(), Some(rd)))
     }
